/*
 * Copyright (c) 2009, INRA All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the INRA nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <vle/extension/Decision.hpp>
#include <vle/utils/Package.hpp>
#include <sstream>
#include <iostream>
#include <fstream>
#include <numeric>
#include <iterator>

#include "CropDvtStade.hpp"

namespace vd = vle::devs;
namespace vv = vle::value;
namespace ved = vle::extension::decision;
namespace vu = vle::utils;

namespace record {
namespace cv {

class DecAgent: public ved::Agent
{
public:
    DecAgent(const vd::DynamicsInit& mdl, const vd::InitEventList& evts) :
        ved::Agent(mdl, evts), mPluies(5, 0.), mStade(BARE_SOIL),
        mDateDebutMaturite(0.0), mRecolteEffectuee(false)
    {
        //initialisation des paramètres
        stockEau = evts.getDouble("stockEau");
        apportEau = evts.getDouble("apportEau");
        quantiteEauSolPortant = evts.getDouble("quantiteEauSolPortant");
        quantiteEauPeriodeSeche = evts.getDouble("quantiteEauPeriodeSeche");

        //inititialisation du KnowledgeBase
        addFacts(this) +=
            F("Rain", &DecAgent::Rain),
            F("DvtStade", &DecAgent::DvtStade);

        addPredicates(this) +=
            P("periodeSeche", &DecAgent::periodeSeche),
            P("eauDisponible", &DecAgent::eauDisponible),
            P("finPeriodeSemis", &DecAgent::finPeriodeSemis),
            P("maturiteAtteinte", &DecAgent::maturiteAtteinte),
            P("solPortant", &DecAgent::solPortant),
            P("maturitePlus30jours", &DecAgent::maturitePlus30jours),
            P("recolteNonEffectuee", &DecAgent::recolteNonEffectuee);

        addOutputFunctions(this) +=
            O("semis", &DecAgent::semis),
            O("irrigation", &DecAgent::irrigation),
            O("recolte", &DecAgent::recolte);

        addAcknowledgeFunctions(this) +=
	        A("ackRecolte",&DecAgent::ackRecolte),
            A("ackIrrigation",&DecAgent::ackIrrigation);

        //Initalisation du plan
        vu::Package pack("tp5_1_correction");
        std::string filePath = pack.getDataFile("DecAgentPlan.txt");;
        std::ifstream fileStream(filePath.c_str());
        KnowledgeBase::plan().fill(fileStream);
    }

    virtual ~DecAgent()
    {
    }

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * Predicats
     * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

    /**
     * @brief predicat periodeSeche
     * (somme de pluie sur les 5 derniers jours < quantiteEauPeriodeSeche)
     */
    bool periodeSeche() const
    {
        return std::accumulate(mPluies.begin(), mPluies.begin() + 5, 0.)
            <= quantiteEauPeriodeSeche;
    }
    /**
     * @brief predicat eauDisponible
     * (le stock d'eau n'est pas vide)
     */
    bool eauDisponible() const
    {
        return stockEau > 0;
    }
    /**
     * @brief predicat finPeriodeSemis
     * (on est a la date de fin d'activite Semis).
     */
    bool finPeriodeSemis() const
    {
        return activite("Semis").finish() == currentTime();
    }
    /**
     * @brief predicat maturiteAtteinte
     * (la variable mDateDebutMaturite a ete initialisee)
     */
    bool maturiteAtteinte() const
    {
        return (mDateDebutMaturite > 0);
    }
    /**
     * @brief predicat solPortant
     * (somme de pluie sur les 3 derniers jours < quantiteEauSolPortant)
     */
    bool solPortant() const
    {
        return std::accumulate(mPluies.begin(), mPluies.begin() + 3, 0.)
        <= quantiteEauSolPortant;
    }
    /**
     * @brief predicat maturitePlus30jours
     * (la date de maturite est depassee exactement de 30 jours)
     */
    bool maturitePlus30jours() const
    {
        return currentTime() == mDateDebutMaturite + 30;
    }

    /**
     * @brief predicat recolteNonEffectuee
     * (à vrai si a recolte n'a pas été effectuée).
     */
    bool recolteNonEffectuee() const
    {
        return not mRecolteEffectuee;
    }

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * Facts
     * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

    /**
     * @brief mise a jour du fait DvtStade
     */
    void DvtStade(const vv::Value& value)
    {
        int newstade = std::floor(value.toDouble().value());
        if (newstade == MATURITY && mStade != MATURITY) {
            mDateDebutMaturite = mCurrentTime;
        }
        mStade = (CROP_STATE) newstade;
    }

    /**
     * @brief mise a jour du fait Rain
     */
    void Rain(const vv::Value& value)
    {
        std::rotate(mPluies.rbegin(), mPluies.rbegin() + 1, mPluies.rend());
        mPluies[0] = value.toDouble().value();
    }

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * Output functions
     * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

    /**
     * @brief activation du semis
     */
    void semis(const std::string& name, const ved::Activity& activity,
            vd::ExternalEventList& output)
    {
        if (activity.isInStartedState()) {
            vd::ExternalEvent* evt = new vd::ExternalEvent("Sowing");
            evt->putAttribute("name", new vv::String(name));
            evt->putAttribute("activity", new vv::String(name));
            evt->putAttribute("value", new vv::Double(1.));
            output.push_back(evt);
        }
    }

    /**
     * @brief activation de la recolte
     */
    void recolte(const std::string& name, const ved::Activity& activity,
            vd::ExternalEventList& output)
    {
        if (activity.isInStartedState()) {
            vd::ExternalEvent* evt = new vd::ExternalEvent("Harvesting");
            evt->putAttribute("name", new vv::String(name));
            evt->putAttribute("activity", new vv::String(name));
            evt->putAttribute("value", new vv::Double(1.));
            output.push_back(evt);
        }
    }

    /**
     * @brief activation d'une irrigation
     */
    void irrigation(const std::string& name, const ved::Activity& activity,
            vd::ExternalEventList& output)
    {
        if (activity.isInStartedState()) {
            vd::ExternalEvent* evt = new vd::ExternalEvent("Irrigation");
            evt->putAttribute("name", new vv::String("Irrigation"));
            evt->putAttribute("activity", new vv::String(name));
            evt->putAttribute("value",
                    new vv::Double(std::min(apportEau, stockEau)));
            output.push_back(evt);
        }
    }

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * Acknowledge functions
     * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
    void ackIrrigation(const std::string&
            /*activityname*/, const ved::Activity& /*activity*/)
    {
        stockEau -= apportEau;
    }

    void ackRecolte(const std::string&
            /*activityname*/, const ved::Activity& /*activity*/)
    {
        mRecolteEffectuee = true;
    }


private:

    //Parametres de l'agent
    double stockEau;
    double apportEau;
    double quantiteEauSolPortant;
    double quantiteEauPeriodeSeche;

    //variables de stockage de données et de
    std::vector<double> mPluies;
    CROP_STATE mStade;
    vd::Time mDateDebutMaturite;
    bool mRecolteEffectuee;

    /**
     * @brief recupere une activite
     * @param nom, nom de l'activite
     * @return l'activite de nom 'nom'
     */
    const ved::Activity& activite(const std::string& nom) const
    {
        return activities().get(nom)->second;
    }

};

}
} // namespaces

DECLARE_DYNAMICS(record::cv::DecAgent)
